'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _utils = require('../utils');

var _looksLike = require('../utils/looks-like');

var _looksLike2 = _interopRequireDefault(_looksLike);

var _identifiers = require('./identifiers');

var _signatures = require('./signatures');

var _signatures2 = _interopRequireDefault(_signatures);

var _templates = require('./templates');

var _templates2 = _interopRequireDefault(_templates);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

exports.default = (path, template, t) => {
  const { buildGroupCollapsed, buildGroupEnd, buildLog } = (0, _templates2.default)(template);

  const scope = traverseFunctions(path, t);

  if (!scope) {
    return;
  }

  const { parameterIdentifiers, parentScope, returnStatement, signature, variableIdentifiers } = scope;

  const mapIdentifer = identifier => {
    const { loc: { start: { column, line } }, name } = identifier;
    return buildLog(t.stringLiteral(`(${line}:${column})`), t.stringLiteral(`${name}:`), identifier);
  };
  const parameters = parameterIdentifiers.map(mapIdentifer);
  const variables = variableIdentifiers.map(mapIdentifer);
  const buildReturn = returnStatement => {
    if (returnStatement) {
      const { loc: { start: { column, line } } } = returnStatement;
      return buildLog(t.stringLiteral(`(${line}:${column})`), returnStatement);
    }
    return buildLog(t.stringLiteral('Void'));
  };
  const scriptScope = Object.keys(parentScope).filter(key => {
    const binding = parentScope[key];
    const isRequire = (0, _looksLike2.default)(binding.path, { node: { init: { callee: { name: 'require' } } } });
    return binding.kind !== 'module' && !isRequire;
  }).map(key => {
    const { identifier } = parentScope[key];
    const { loc: { start: { column, line } }, name } = identifier;
    return buildLog(t.stringLiteral(`(${line}:${column})`), t.stringLiteral(`${name}:`), identifier);
  });

  return [buildLog(...path.node.arguments), buildGroupCollapsed(t.stringLiteral(signature)), buildGroupCollapsed(t.stringLiteral('Parameters')), ...parameters, buildGroupEnd(t.stringLiteral('Parameters')), buildGroupCollapsed(t.stringLiteral('Variables')), ...variables, buildGroupEnd(t.stringLiteral('Variables')), buildGroupCollapsed(t.stringLiteral('Return')), buildReturn(returnStatement), buildGroupEnd(t.stringLiteral('Return')), buildGroupCollapsed(t.stringLiteral('Script')), ...scriptScope, buildGroupEnd(t.stringLiteral('Script')), buildGroupEnd(t.stringLiteral(signature))];
};

const traverseFunctions = (path, t) => {
  const parentFunction = findParentFunction(path, t);

  if (!parentFunction) {
    return null;
  }

  let rootFunction = parentFunction;
  let variables = [(0, _identifiers.getVariableIdentifiers)(parentFunction.scope.bindings)];
  let parameters = [(0, _identifiers.getParamIdentifiers)(parentFunction.scope.bindings)];

  while (isNotRootFunction(rootFunction, t)) {
    const maybeParentFunction = findParentFunction(rootFunction, t);
    if (!maybeParentFunction) {
      break;
    }
    rootFunction = maybeParentFunction;
    variables.unshift((0, _identifiers.getVariableIdentifiers)(rootFunction.scope.bindings));
    parameters.unshift((0, _identifiers.getParamIdentifiers)(rootFunction.scope.bindings));
  }

  const parentScope = rootFunction.scope.parent.bindings;
  const stmt = parentFunction.node.body.body.find(rest => t.isReturnStatement(rest));
  const returnStatement = stmt && stmt.argument;

  return {
    parameterIdentifiers: (0, _utils.flatten)(parameters),
    parentScope,
    returnStatement,
    signature: (0, _signatures2.default)(t, parameters, rootFunction),
    variableIdentifiers: (0, _utils.flatten)(variables)
  };
};

const findParentFunction = (path, t) => path.findParent(parent => {
  return t.isFunctionDeclaration(parent) || t.isArrowFunctionExpression(parent) || t.isFunctionExpression(parent);
});

const isNotRootFunction = (path, t) => !(0, _looksLike2.default)(path, { node: { id: t.isIdentifier } }) && !t.isVariableDeclarator(path.parent) && !t.isFunctionExpression(path.parent);